// BeOS driver for Alcatel Speed Touch USB ADSL modem
// Philippe Houdoin
//
// References
// ----------
//
// Linux Speed Touch USB open source driver
//
// DONE
// ----
//
// - USB detection
//
// TODO
// ----
// - all!
//
// HISTORY
// -------
// March 28, 2001		start this project (speed_touch)

#include <stdio.h>

#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <ByteOrder.h>
#include <USB.h>

#define DEBUG 1
#include <Debug.h>

#include "driver.h"
#include "speed_touch_usb.h"	// Alcatel Speed Touch USB related definitions


#ifdef CODEWARRIOR
	#pragma mark [Local definitions]
#endif

// Local definitions
// -----------------

#ifndef DRIVER_NAME
	#define DRIVER_NAME 		"generic"
#endif
#define DEBUGGER_COMMAND	 	DRIVER_NAME

#define DEVICE_NAME 			DRIVER_NAME
#define DEVICE_NAME_LENGTH		(64)			
#define DEVICE_PATH 			"net/speed_touch/%d"

typedef struct
	{
	uint8	 byte[6];
	}
mac_address_t;

typedef struct device device;

struct device {
	device *					next;				// linked list
	
	long						id;
	char 						name[DEVICE_NAME_LENGTH + 1];
	uint32						flags;
	uint32						debug_level;

	// ---- Locks mechanisms
	uint32						open_count;
	
	sem_id						event_semaphore;
	sem_id						read_semaphore;
	sem_id						write_semaphore;		

	// ---- device data
	const usb_device *			ud;						// our usb device
/*
	const usb_interface_info *	comm_interface;			// ACM communication interface
	const usb_endpoint_info *	control_int_endpoint;	// Control interrupt endpoint

	const usb_interface_info *	data_interface;			// ACM data interface
	const usb_endpoint_info *	bulk_in_endpoint;		// Data bulk in endpoint
	const usb_endpoint_info *	bulk_out_endpoint;		// Data bulk out endpoint
*/	
	// ---- device buffers
	void *						buffer;					// Common buffer (for interrupt, bulk in and bukl out)
	area_id						buffer_area_id;			// Common buffer area id

	void *						control_buffer;			// Address of control data in common buffer
	void *						write_buffer;			// Address of write data in common buffer
	uint32						write_status;
	uint32						write_actual_length;	
	void *						read_buffer;			// Address of read data in common buffer
	uint32						read_status;
	uint32						read_actual_length;
	uint32						unread_bytes;

	// ---- device states
	mac_address_t				mac_address;	// MAC address
	
	// ---- Somes statistics
	volatile					nb_bytes_sent;
	volatile					nb_bytes_received;
	volatile					nb_request_failure;
	volatile					nb_write_errors;
	volatile					nb_read_errors;
};

// globals for driver instances
static usb_module_info *	g_usb = NULL;
static char *				g_driver_name = DRIVER_NAME;

// devices list of devices instances (single linked list)
static sem_id				g_device_list_lock = -1;		// devices list locker 
static device *	 			g_device_list = NULL;			// head pointeur, see 'next' in device type 
static char **				g_device_name_list = NULL;		// dynamically compute in publish_devices()


// Prototypes of device hooks functions
static status_t device_added(const usb_device * usb_dev, void ** cookie);
static status_t device_removed(void * cookie);

static status_t device_open(const char *name, uint32 flags, void ** cookie);
static status_t device_close(void * cookie);
static status_t device_free(void * cookie);
static status_t device_control(void * cookie, uint32 msg,void * buf, size_t len);
static status_t device_read(void * cookie, off_t pos, void * buf, size_t * len);
static status_t device_write(void * cookie, off_t pos, const void * buf, size_t * len);

/* And, so, the device hooks list */
static device_hooks g_device_hooks = {
	device_open, 			/* -> open entry point */
	device_close,         	/* -> close entry point */
	device_free,          	/* -> free entry point */
	device_control,       	/* -> control entry point */
	device_read,       	  	/* -> read entry point */
	device_write,        	/* -> write entry point */
	NULL,               	/* -> select entry point */
	NULL,               	/* -> deselect entry point */
	NULL,               	/* -> readv */
	NULL                	/* -> writev */
};

// For USB module
static usb_notify_hooks g_usb_notify_hooks = {
	device_added,
	device_removed
};

static usb_support_descriptor g_usb_supported_devices[] = {
	{ 0, 0, 0, ALCATEL_VENDOR_ID, SPEED_TOUCH_PRODUCT_ID}
};

#ifdef CODEWARRIOR
	#pragma mark [Local prototypes]
#endif

// Privates prototypes
// -------------------

static long			device_count();
static device * 	device_find_by_name(const char * name);
static status_t 	device_init(device * dev);
static status_t		device_uninit(device * dev);

#ifdef DEBUGGER_COMMAND
	static int 		device_debugger_command(int argc, char ** argv);
#endif

static status_t 	device_clear_statistics(device * dev);

#define ROUND_UP(x, y) (((x) + (y) - 1) & ~((y) - 1))


// OKAY, NOW IMPLEMENTATION PLEASE!
// --------------------------------

#ifdef CODEWARRIOR
	#pragma mark [Driver API calls]
#endif

// Driver API calls
// ----------------


_EXPORT int32 api_version = B_CUR_DRIVER_API_VERSION;


// -------------------------------------------------------
_EXPORT status_t init_hardware (void)
{
	DEBUG_ONLY( dprintf(LOGID "--- init_hardware\n") );
	return B_OK;
}


// -------------------------------------------------------
_EXPORT status_t init_driver(void)
{
	status_t	status;

	DEBUG_ONLY( dprintf(LOGID "--- init_driver, built %s %s\n", __DATE__, __TIME__) );

#ifdef DEBUG
	if ( load_driver_symbols(DRIVER_NAME) != B_OK )
		dprintf(LOGID WARN "load_driver_symbols(\"" DRIVER_NAME "\") FAILED !\n");
	else
		dprintf(LOGID "load_driver_symbols(\"" DRIVER_NAME "\") done.\n");
#endif

	status = get_module(B_USB_MODULE_NAME, (module_info **) &g_usb);
	if ( status != B_OK )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "get_module(\"" B_USB_MODULE_NAME "\") FAILED: %s\n",
							strerror(status)) );
		return status;
		};

	if ((g_device_list_lock = create_sem (1, DRIVER_NAME " device_list_lock")) < 0)
		{
		put_module (B_USB_MODULE_NAME);
		return g_device_list_lock;		// error code
		};

	set_sem_owner(g_device_list_lock, B_SYSTEM_TEAM);
		
	g_usb->register_driver(g_driver_name, g_usb_supported_devices, 1, NULL);
	g_usb->install_notify(g_driver_name, &g_usb_notify_hooks);
	
#ifdef DEBUGGER_COMMAND
	add_debugger_command(DEBUGGER_COMMAND, device_debugger_command, DRIVER_NAME " driver info");
#endif	

	return B_OK;
}


// -------------------------------------------------------
_EXPORT void uninit_driver(void)
{
	long 		i;
	device *	dev;
	device * 	next_dev;

	DEBUG_ONLY( dprintf(LOGID "--- uninit_driver\n") );

	if ( g_usb != NULL )
		g_usb->uninstall_notify(g_driver_name);
	g_usb = NULL;
	
	put_module(B_USB_MODULE_NAME);

#ifdef DEBUGGER_COMMAND
	remove_debugger_command(DEBUGGER_COMMAND, device_debugger_command);
#endif

	// Free device list
	dev = g_device_list;
	while ( dev )
		{
		next_dev = dev->next;
		
		device_uninit(dev);
		free(dev);
		
		dev = next_dev;
		};

	// Free device name list
	if( g_device_name_list )
		{
		for( i = 0; g_device_name_list[i]; i++ )
			free((char *) g_device_name_list[i]);
		free(g_device_name_list);
		};
	
	delete_sem(g_device_list_lock);
}


// -------------------------------------------------------
_EXPORT const char ** publish_devices(void)
{
	long 	i;

	DEBUG_ONLY( dprintf(LOGID "--- publish_devices\n") );

	// free previous device names list
	if( g_device_name_list )
		{
		for( i = 0; g_device_name_list[i]; i++ )
			free((char *) g_device_name_list[i]);
		free(g_device_name_list);
		};
		
	// build device names list
	acquire_sem(g_device_list_lock);	
	g_device_name_list = (char **) malloc(sizeof(char*) * (device_count() + 1));	
	if( g_device_name_list )
		{
		device * dev;
		
		// iterate each device in list, and build his name...
		i = 0;
		dev = g_device_list;
		while ( dev )
			{
			g_device_name_list[i] = (char *) malloc(strlen(dev->name) + 1);
			if( g_device_name_list[i] )
				{
				strcpy(g_device_name_list[i], dev->name);
				DEBUG_ONLY( dprintf(LOGID "%s\n", g_device_name_list[i]) );
				i++;
				};
			
			// next device
			dev = dev->next;
			};

		// end the device names list
		g_device_name_list[i] = NULL;
		};

	release_sem(g_device_list_lock);
	
	return (const char **) g_device_name_list;
}


// -------------------------------------------------------
_EXPORT device_hooks * find_device
	(
	const char *	name
	)
{
	DEBUG_ONLY( dprintf(LOGID "--- find_device %s\n", name) );
	return &g_device_hooks;
}


#ifdef CODEWARRIOR
	#pragma mark [Device hooks]
#endif

// Device hooks
// ------------

// -------------------------------------------------------
static status_t device_added
	(
	const usb_device *	ud,
	void **				cookie
	)
{
	const usb_configuration_info * 	uci;
	const usb_interface_info *		uii;
	const usb_interface_info *		comm_interface;
	const usb_endpoint_info *		control_int_endpoint;
	const usb_interface_info *		data_interface;
	const usb_endpoint_info *		bulk_in_endpoint;
	const usb_endpoint_info *		bulk_out_endpoint;
	device *						dev;
	status_t						status;
	int								c;
	int								i;
	int								alt;
	char							resource_name[B_OS_NAME_LENGTH + 1];
		
	
	DEBUG_ONLY( dprintf(LOGID "--- device_added\n") );

	dev = NULL;

	if ( NULL == g_usb )
		return B_ERROR;

	status = B_ERROR;
		
	// look up for Speed Touch USB (or what seem to be one...) configuration
	c = 0;
	while ( true )	// I *know* ! A infinite loop !!! Here !!! In a driver code !!!
		{			// But, relax, there's break condition(s).
		uci = g_usb->get_nth_configuration(ud, c);
		c++;		// Be prepared for next configuration (for *continue* statement)
		if ( NULL == uci )
			break;			// First *break* condition: no more configuration
/*
		status = g_usb->set_alt_interface(ud, uii);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(LOGID ERR "set_alt_interface(%d) FAILED (%d)!\n",
						uii->descr->interface_number, status) );
			continue;	// Maybe next configuration is the one?
			};
*/



		status = g_usb->set_configuration(ud, uci);
		if ( status != B_OK )
			{
			// Can't set this configuration !!!
			DEBUG_ONLY( dprintf(LOGID ERR "set_configuration(%d) FAILED (%d)!\n", c, status) );
			continue;	// Maybe next configuration is the one?
			};

		dev = (device *) malloc(sizeof(*dev) + 4);
		if ( dev == NULL )
			{
			// Can't allocate memory for this device!!!
			// We abort, because even if another configuration can be a good one,
			// we hardly would success to allocate again, so why persist??? 
			DEBUG_ONLY( dprintf(LOGID ERR "Can't allocate(device). ABORT!\n") );
			status = B_NO_MEMORY;
			goto failed;
			};
	
		// Init this device block/cookie with invalid values, for failure exit
		memset(dev, 0, sizeof(*dev));

		dev->open_count = 0;
		
		dev->event_semaphore 	= -1;
		dev->read_semaphore 	= -1;
		dev->write_semaphore 	= -1;

		dev->buffer_area_id 	= -1;

		// Okay, setup this device block/cookie		
		dev->ud = ud;

#if DEBUG
	{
	size_t					sz;
	void *					buffer;
	usb_device_descriptor *	udd;
	usb_string_descriptor *	usd;
	uint8					manufacturer = 0;
	uint8					product = 0;
	uint8					serial_number = 0;

	sz = 512;
	buffer = malloc(sz);
	memset(buffer, 0, sz);
		
	status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_DEVICE, 0, 0, buffer, sz, &sz);
	if (status != B_OK)
		dprintf(LOGID ERR "get_descriptor(USB_DESCRIPTOR_DEVICE) FAILED (%d)!\n", status);
	else
		{
		udd = buffer;
		manufacturer 	= udd->manufacturer;
		product			= udd->product;
		serial_number	= udd->serial_number;
		};

	sz = 512;
	memset(buffer, 0, sz);

	#define USB_LANGIDS	(0)

	status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING, USB_LANGIDS, 0, buffer, sz, &sz);
	if (status != B_OK)
		dprintf(LOGID ERR "get_descriptor(USB_LANGIDS) FAILED (%d)!\n", status);
	else
		{
		uint16 *	langid;
		int			i;
	
		usd = buffer;
		langid = (uint16 *) usd->string;
		
		dprintf(LOGID " LANGIDS supported:\n");
		for( i = 0; i < ((usd->length - 2) / 2); i++)
			{
			dprintf(LOGID " 0x%x\n", *langid);
			langid++;
			};
		};

	if ( manufacturer )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING, manufacturer, 0x0409, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(manufacturer) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			dprintf(LOGID " manufacturer: [%s]\n", text);
			};
		};

	if ( product )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING, product, 0x0409, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(product) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			dprintf(LOGID " product: [%s]\n", text);
			};
		};

	if ( serial_number )
		{
		sz = 512;
		memset(buffer, 0, sz);

		status = g_usb->get_descriptor(ud, USB_DESCRIPTOR_STRING, serial_number, 0x0409, buffer, sz, &sz);
		if (status != B_OK)
			dprintf(LOGID ERR "get_descriptor(manufacturer) FAILED (%d)!\n", status);
		else
			{
			char 	text[256];
			int i;
			
			usd = buffer;
			for (i=0; i < (usd->length/2); i++)
				text[i] = usd->string[2*i];	// UNICODE string
			text[i] = 0;
			dprintf(LOGID " serial_number: [%s]\n", text);
			
#define HEX2INT(c) ( (c >= '0')&&(c <= '9') ?  (c - '0') : ((c & 0xf)+9) )

			for (i=0; i<6; i++)
			    dev->mac_address.byte[i] = (HEX2INT(text[i*2])*16) + (HEX2INT(text[i*2+1]));

			dprintf(LOGID " MAC address: %02x:%02x:%02x:%02x:%02x:%02x\n",
					dev->mac_address.byte[0], dev->mac_address.byte[1],
					dev->mac_address.byte[2], dev->mac_address.byte[3],
					dev->mac_address.byte[4], dev->mac_address.byte[5] );

			};
		};

	free(buffer);
	}
#endif

		// Compute device unique id
		acquire_sem(g_device_list_lock);
		dev->id = device_count();
		release_sem(g_device_list_lock);
		sprintf(dev->name, DEVICE_PATH, dev->id);

		// -- Okay, okay okay, well done!

		// Add this new device to device (linked) list 					
		acquire_sem(g_device_list_lock);
		dev->next = g_device_list;	// No list order, so we add new device at list head
		g_device_list = dev;
		release_sem(g_device_list_lock);



		*cookie = dev;
		return B_OK;	// Second *break* condition: device fully detected
		};	// while ( configuration )...

failed:;

	if ( dev != NULL )
		{

		};

	// Glup, this device isn't a (at least supported) Speed Touch USB device
	return status;
}


// -------------------------------------------------------
static status_t device_removed
	(
	void *			cookie
	)
{
	device *	dev;
	
	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_removed: COOKIE BURNED !\n") );
		return EINVAL;
		};


	// remove it from the list of devices	
	acquire_sem(g_device_list_lock);
	if(dev == g_device_list)
		{
		g_device_list = dev->next;
		}
	else
		{
		device * tmp;
		tmp = g_device_list;
		while ( tmp )
			{
			if ( tmp->next == dev )
				{
				// remove this device from the device (linked) list
				tmp->next = dev->next;
				break;
				};
			tmp = tmp->next;
			};
		};
	release_sem(g_device_list_lock);
	
	// Okay, now try to free device
	if ( atomic_add(&(dev->open_count), -1) == 0 )
		{
		// not (still) open, we can free this device now
		device_uninit(dev);
		DEBUG_ONLY( dprintf(LOGID "--- device_removed %s\n", dev->name) );
		free(dev);
		}
	else
		// still open, device_free hook will free this device
		DEBUG_ONLY( dprintf(LOGID "--- device_removed: %s still open, marked for removal\n", dev->name) );
	
	return B_OK;
}


// -------------------------------------------------------
static status_t device_open
	(
	const char *	name, 
	uint32 			flags, 
	void **			cookie
	)
{
	device *	dev;
	status_t	status;

	dev = device_find_by_name(name);
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_open: %s NOT FOUND, ARGH !\n", name) );
		return EINVAL;
		};
		
	if ( atomic_add(&(dev->open_count), 1) == 0 )	// atomic_xxx() return previous value
		{
		// First open on this device :
		status = device_init(dev);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(LOGID ERR "--- device_open: device_init(%s)->%d, FAILURE !\n", dev->name, status) );
			return status;
			};
		}
	else
		{
		// device already open! sorry
		DEBUG_ONLY( dprintf(LOGID "--- device_open: %s already opened, sorry.\n", dev->name) );
		return B_BUSY;
		};

	dev->flags = flags;

	DEBUG_ONLY( dprintf(LOGID "--- device_open %s (%s%s)\n",
						dev->name,
						( ((dev->flags & O_RWMASK) == O_RDONLY) ? "O_RDONLY" :
						  ((dev->flags & O_RWMASK) == O_WRONLY) ? "O_WRONLY" : "O_RDWR"),
						(dev->flags & O_NONBLOCK) ? " O_NONBLOCK" : "") );
						
	*cookie = dev;

	return B_OK;
}


// -------------------------------------------------------
static status_t device_close
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_close: COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(LOGID "--- device_close %s\n", dev->name) );

	atomic_add(&(dev->open_count), -1);
	
	return B_OK;
}


// -------------------------------------------------------
static status_t device_free
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_free: COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(LOGID "--- device_free %s\n", dev->name) );
	
	// device_uninit(dev);
	// free(dev);

	return B_OK;
}



// -------------------------------------------------------
static status_t device_read
	(
	void *		cookie, 
	off_t 		pos, 
	void *		data, 
	size_t *	len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_read: COOKIE BURNED !\n") );
		return B_ERROR;
		};

	if ( (dev->flags & O_RWMASK) != O_RDONLY && (dev->flags & O_RWMASK) != O_RDWR )
		// Device not open with READ permission!!!
		return B_PERMISSION_DENIED;

	DEBUG_ONLY( dprintf(LOGID "--- device_read %s: %x, %d\n",
				dev->name, data, *len ) );

	*len = 0;

	status = B_UNSUPPORTED;
	return status;
}


// -------------------------------------------------------
static status_t device_write
	(
	void *			cookie, 
	off_t 			pos, 
	const void *	data, 
	size_t *		len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_write: COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	if ( (dev->flags & O_RWMASK) != O_WRONLY && (dev->flags & O_RWMASK) != O_RDWR )
		// Device not open with WRITE permission!!!
		return B_PERMISSION_DENIED;

	DEBUG_ONLY( dprintf(LOGID "--- device_write %s: %x, %d\n",
				dev->name, data, *len ) );

	*len = 0;

	status = B_UNSUPPORTED;
	return status;
}


/*
 * ioctls: belongs in a public header file
 * somewhere, so that the net_server and other ethernet drivers can use.
 */
enum {
	ETHER_GETADDR = B_DEVICE_OP_CODES_END,	/* get ethernet address */
	ETHER_INIT,								/* set irq and port */
	ETHER_NONBLOCK,							/* set/unset nonblocking mode */
	ETHER_ADDMULTI,							/* add multicast addr */
	ETHER_REMMULTI,							/* rem multicast addr */
	ETHER_SETPROMISC,						/* set promiscuous */
	ETHER_GETFRAMESIZE						/* get frame size */
};


// -------------------------------------------------------
static status_t device_control
	(
	void *	cookie, 
	uint32 	op_code, 
	void *	data, 
	size_t 	len
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(LOGID ERR "--- device_control: COOKIE BURNED !\n") );
		return B_ERROR;
		};

	switch ( op_code ) {
	case ETHER_NONBLOCK:
		{
		// Set to blocking i/o on/off
		if ( *((bool *) data) )
			atomic_or(&(dev->flags), (uint32) O_NONBLOCK);
		else
			atomic_and(&(dev->flags), (uint32) ~O_NONBLOCK);
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: ETHER_NONBLOCK -> %s\n",
					dev->name, (dev->flags & O_NONBLOCK) ? "on" : "off") );

		return B_NO_ERROR;
		};

	case ETHER_INIT:
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: ETHER_INIT\n", dev->name) );
		
		return B_OK;
		};
			
	case ETHER_GETADDR:
		{
		uint8 	i;

		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: ETHER_GETADDR\n", dev->name) );
		for ( i = 0; i < 6; i++ )
			((uint8 *) data)[i] = dev->mac_address.byte[i];
		return B_NO_ERROR;
		};

	default:
		{
		DEBUG_ONLY( dprintf(LOGID WARN "--- device_control %s: op_code %d (0x%x) unhandled, , data = 0x%x, len=%d.\n",
							dev->name, op_code, op_code, data, len) );
		return EINVAL;
		};
	}
	return B_OK;
}


#ifdef CODEWARRIOR
	#pragma mark [USB callbacks]
#endif

// USB callbacks
// -------------

#ifdef CODEWARRIOR
	#pragma mark [Privates routines]
#endif

// Privates routines
// -----------------

// -------------------------------------------------------
static long device_count()
{
	device *	dev;
	long		count;
	
	count = 0;
	dev = g_device_list;
	while ( NULL != dev )
		{
		count++;
		dev = dev->next;
		};

	return count;
}

// -------------------------------------------------------
static device * device_find_by_name
	(
	const char *	name
	)
{
	device *	dev;
	
	acquire_sem(g_device_list_lock);
	dev = g_device_list;
	while ( dev )
		{
		if ( strcmp(dev->name, name) == 0 )
			{
			// name found!
			release_sem(g_device_list_lock);
			return dev;
			};

		// next one, maybe?
		dev = dev->next;
		};

	release_sem(g_device_list_lock);	
	return NULL;
}

// -------------------------------------------------------
static status_t device_init
	(
	device *	dev
	)
{
	status_t		status;

	status = B_ERROR;
	
	if ( NULL == dev )
		return B_BAD_VALUE;
		
	DEBUG_ONLY( dprintf(LOGID "--- device_init %s\n", dev->name) );

	// init startup values
	dev->flags					= 0;

	dev->event_semaphore		= -1;
	dev->read_semaphore			= -1;
	dev->unread_bytes			= 0;
	dev->write_semaphore		= -1;

	device_clear_statistics(dev);

	status = B_OK;
	
	return status;
}


// -------------------------------------------------------
static status_t device_uninit
	(
	device *	dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	DEBUG_ONLY( dprintf(LOGID "--- device_uninit %s\n", dev->name) );

	delete_sem(dev->event_semaphore);
	delete_sem(dev->read_semaphore);
	delete_sem(dev->write_semaphore);

	return B_OK;
}



// -------------------------------------------------------
static status_t device_clear_statistics
	(
	device * dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	dev->nb_bytes_sent 		= 0;
	dev->nb_bytes_received 	= 0;
	dev->nb_request_failure	= 0;
	dev->nb_write_errors	= 0;
	dev->nb_read_errors		= 0;
	
	return B_OK;
};


#ifdef CODEWARRIOR
	#pragma mark [Debugger Command]
#endif


#ifdef DEBUGGER_COMMAND

// -------------------------------------------------------
static int device_debugger_command
	(
	int 	argc,
	char **	argv
	)
{
	struct debugger_commands
		{
		int		id;
		char * 	name;
		char *	shortcut;
		// TO DO : command handling by sub-routine call
		// int 	(*cmd)(device * dev, int argc, char ** argv);
		char *	arguments;
		}
	* cmd, dbgr_cmds[] =
		{
			{0,  "help", "h", NULL},
			{1,  "list", "l", NULL},
			{2,  "select", "s", "{index}"},
			{3,  "stats", "st", NULL},
			{4,  "clear_stats", "cs", NULL},
			{5,  "lockers", "lck", NULL},
			{6,	 "dump_device", "dd", NULL},
			{0,  NULL, NULL, NULL}
		};
	int 			cmd_id;
	int				return_value;
	int				i;
	
	static int		dev_id = -1;
	device *		dev;

	return_value 	= 0;

	i = 0;
	dev = g_device_list;
	while ( dev )
		{
		if ( i == dev_id )
			break;
		i++;
		dev = dev->next;
		};

	if ( NULL == dev )
		kprintf("  no device selected !\n");
	else
		NULL; // w = dev->data;

	cmd_id 			= 0;	// help, by default

	if ( argc > 1 )
		{
		cmd = dbgr_cmds;
		while ( cmd->name )
			{
			if ( (strcmp(cmd->name, argv[1])== 0) ||
				 (strcmp(cmd->shortcut, argv[1]) == 0 ) )
				break;

			 cmd++;
			};

		cmd_id = cmd->id;
		};

	switch (cmd_id)
		{
		case 1: // list
			{
			device *	tmp_dev;

			i = 0;
			tmp_dev = g_device_list;
			while ( tmp_dev )
				{
				kprintf("  %d: %s", i, tmp_dev->name);
				if ( i == dev_id )
					kprintf(" (selected)");
				kprintf("\n");

				i++;
				tmp_dev = tmp_dev->next; 
				};
			break;
			};

		case 2:	// select
			{
			if ( argc >= 3 )
				{
				long	count;
				
				i = atoi(argv[2]);
				count = device_count();
				if ( (i < 0) || (i >= count) )
				 	kprintf("  index must be in [0, %d] range !\n", count - 1);
				else
					dev_id = i;
				}
			else
				kprintf("  index missing !\n");
			break;
			};
			
		case 3:	// stats
			{
			if ( NULL == dev )
				break;

		 	kprintf("  nb_bytes_sent		%d\n"
					"  nb_bytes_received	%d\n"
					"  nb_request_failure	%d\n"
					"  nb_write_errors		%d\n"
					"  nb_read_errors		%d\n",
					dev->nb_bytes_sent, dev->nb_bytes_received, dev->nb_request_failure,
					dev->nb_write_errors, dev->nb_read_errors);
			break;
			};
			
		case 4:	// clear stats
			{
			if ( NULL == dev )
				break;
			device_clear_statistics(dev);
			break;
			};

		case 5: // locker
			{
			if ( NULL == dev )
				break;

			kprintf("  device at 0x%4.4x :\n"
			        "    id                %d\n"
					"    name              %s\n"
					"    open_count        %d\n"
					"    flags             %x\n",
					dev, dev->id, dev->name, dev->open_count, dev->flags);

			DEBUG_ONLY( kprintf("    debug_level       0x%x\n", dev->debug_level) );
			break;
			};

		case 6: // dump_device
			{
			if ( NULL == dev )
				break;

			kprintf("  device at 0x%4.4x :\n"
			        "    id                %d\n"
					"    name              %s\n"
					"    open_count        %d\n"
					"    flags             %x\n",
					dev, dev->id, dev->name, dev->open_count, dev->flags);

			DEBUG_ONLY( kprintf("    debug_level       0x%x\n", dev->debug_level) );
			break;
			};
			
		case 0:	// help
		default:
			{
			kprintf(DEBUGGER_COMMAND " usage:\n");
			cmd = dbgr_cmds;
			while ( cmd->name )
				{
				kprintf("  %s", cmd->name);
				if ( cmd->shortcut )
					kprintf(" | %s", cmd->shortcut);
				if ( cmd->arguments )
					kprintf(" %s", cmd->arguments);
				kprintf("\n");

				cmd++;	// next
				};
			
			break;
			};
		};	// switch
	
	return return_value;
}

#endif // DEBUGGER_COMMAND





